package com.ezet.crowdstream.client.streamactivity

import android.Manifest
import android.content.ComponentName
import android.content.Context
import android.content.Intent
import android.content.ServiceConnection
import android.content.pm.PackageManager
import android.location.Location
import android.net.Uri
import android.os.AsyncTask
import android.os.Build
import android.os.Bundle
import android.os.IBinder
import android.preference.PreferenceManager
import android.provider.Settings
import android.support.annotation.RequiresApi
import android.support.v4.app.ActivityCompat
import android.support.v4.content.ContextCompat
import android.util.Log
import android.view.Gravity
import android.view.Menu
import android.view.MenuItem
import android.view.View
import android.view.ViewGroup
import android.widget.FrameLayout
import android.widget.Toast
import com.ezet.crowdstream.client.R
import com.ezet.crowdstream.client.SettingsActivity
import com.ezet.crowdstream.client.services.CwsNetworkService
import com.ezet.crowdstream.client.services.PreferenceService
import com.ezet.crowdstream.client.streamactivity.tasks.DisplaySequenceTask
import com.google.android.gms.location.FusedLocationProviderClient
import com.google.android.gms.location.LocationServices
import io.reactivex.android.schedulers.AndroidSchedulers
import io.reactivex.disposables.Disposable
import kotlinx.android.synthetic.main.activity_stream.*
import java.lang.ref.WeakReference

internal class StreamActivity : BaseFullscreenActivity() {

    private lateinit var preferenceService: PreferenceService
    private var connectionStatusToast: Toast? = null
    private val networkServiceConnection = NetworkServiceConnection(this)
    private var connected = false
    private lateinit var locationProvider: FusedLocationProviderClient
    private var location: Location? = null
    private var streaming = false

    companion object {
        @Suppress("unused")
        private val TAG = "StreamActivity"
        private val PERMISSION_REQUEST_WRITE_SETTINGS: Int = 1
        private val PERMISSION_REQUEST_ACCESS_COARSE_LOCATION: Int = 2
        private val SPL = 2
        private val MAX_BRIGHTNESS = 255
    }

    override fun onCreate(savedInstanceState: Bundle?) {
        super.onCreate(savedInstanceState)
        setContentView(R.layout.activity_stream)
        preferenceService = PreferenceService(application)
        locationProvider = LocationServices.getFusedLocationProviderClient(this)
        color_surface.setOnClickListener { toggle() }
        requestLocationAndConnect()
        requestAndSetBrightness()
        setDefaultPreferences()
    }

    override fun onResume() {
        super.onResume()
        requestAndSetBrightness()
        val (width, height) =
                if (!preferenceService.getBoolean(R.string.pkey_display_fullscreen)) {
                    Pair(300, 300)
                } else {
                    Pair(ViewGroup.LayoutParams.MATCH_PARENT, ViewGroup.LayoutParams.MATCH_PARENT)
                }
        color_surface.layoutParams = FrameLayout.LayoutParams(width, height, Gravity.CENTER)
        if (preferenceService.getBoolean(R.string.pkey_auto_connect)) {
            connect()
        }
    }

    override fun onPause() {
        super.onPause()
        restoreBrightness()
    }

    override fun onDestroy() {
        super.onDestroy()
        unbindService(networkServiceConnection)
        restoreBrightness()
    }

    private fun setDefaultPreferences() {
        val sp = PreferenceManager.getDefaultSharedPreferences(this)
        if (sp.getInt("spl", 0) != Companion.SPL) {
            PreferenceManager.setDefaultValues(this, R.xml.pref_network, true)
            PreferenceManager.setDefaultValues(this, R.xml.pref_display, true)
            sp.edit().putInt("spl", Companion.SPL).apply()
        }
    }

    private fun setMaxBrightness() {
        Log.d(TAG, "setMaxBrightness")
        Settings.System.putInt(
                contentResolver,
                Settings.System.SCREEN_BRIGHTNESS_MODE,
                Settings.System.SCREEN_BRIGHTNESS_MODE_MANUAL
        )
        Settings.System.putInt(
                contentResolver,
                Settings.System.SCREEN_BRIGHTNESS,
                Companion.MAX_BRIGHTNESS
        )
    }

    private fun restoreBrightness() {
        Settings.System.putInt(
                contentResolver,
                Settings.System.SCREEN_BRIGHTNESS_MODE,
                Settings.System.SCREEN_BRIGHTNESS_MODE_AUTOMATIC
        )
    }

    private fun requestLocationAndConnect() {
        val hasPermission: Boolean = ContextCompat.checkSelfPermission(
                this,
                Manifest.permission.ACCESS_COARSE_LOCATION
        ) == PackageManager.PERMISSION_GRANTED
        if (hasPermission) {
            locationProvider.lastLocation.addOnSuccessListener { location = it; connect() }
        } else {
            requestPermission(
                    Manifest.permission.ACCESS_COARSE_LOCATION,
                    PERMISSION_REQUEST_ACCESS_COARSE_LOCATION
            )
        }
    }

    private fun requestAndSetBrightness() {
        val hasPermission: Boolean = if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.M) {
            Settings.System.canWrite(this)
        } else {
            ContextCompat.checkSelfPermission(
                    this,
                    Manifest.permission.WRITE_SETTINGS
            ) == PackageManager.PERMISSION_GRANTED
        }
        if (hasPermission) {
            setMaxBrightness()
        } else {
            if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.M) {
                requestPermissionIntent(
                        Settings.ACTION_MANAGE_WRITE_SETTINGS,
                        Companion.PERMISSION_REQUEST_WRITE_SETTINGS
                )
            } else {
                requestPermission(
                        Manifest.permission.WRITE_SETTINGS,
                        PERMISSION_REQUEST_WRITE_SETTINGS
                )
            }
        }
    }

    private fun requestPermission(permission: String, request_id: Int) {
        if (ActivityCompat.shouldShowRequestPermissionRationale(this, permission)) {
            // show explanation
            ActivityCompat.requestPermissions(
                    this,
                    arrayOf(permission), request_id
            )
        } else {
            ActivityCompat.requestPermissions(
                    this,
                    arrayOf(permission), request_id
            )
        }
    }

    @RequiresApi(Build.VERSION_CODES.M)
    private fun requestPermissionIntent(action: String, request_id: Int) {
        val intent = Intent(action, Uri.parse("package:" + packageName))
        startActivityForResult(intent, request_id)
    }

    override fun onActivityResult(requestCode: Int, resultCode: Int, data: Intent?) {
        when (requestCode) {
            PERMISSION_REQUEST_WRITE_SETTINGS -> {
                if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.M && Settings.System.canWrite(this)) {
                    Log.d(TAG, "CODE_WRITE_SETTINGS_PERMISSION success")
                    setMaxBrightness()
                }
            }
            else -> super.onActivityResult(requestCode, resultCode, data)
        }
    }

    override fun onRequestPermissionsResult(
        requestCode: Int,
        permissions: Array<String>, grantResults: IntArray
    ) {
        when (requestCode) {
            Companion.PERMISSION_REQUEST_WRITE_SETTINGS -> {
                if (grantResults.isNotEmpty() && grantResults[0] == PackageManager.PERMISSION_GRANTED) {
                    Log.i(TAG, "Permission granted")
                    setMaxBrightness()
                } else {
                    Log.i(TAG, "Permission denied")
                }
                return
            }
            Companion.PERMISSION_REQUEST_ACCESS_COARSE_LOCATION -> {
                if (grantResults.isNotEmpty() && grantResults[0] == PackageManager.PERMISSION_GRANTED) {
                    requestLocationAndConnect()
                }
            }
        }
    }

    private fun bindNetworkService() {
        val intent = Intent(this, CwsNetworkService::class.java)
        bindService(intent, networkServiceConnection, Context.BIND_AUTO_CREATE)
    }

    private fun connect() {
        if (networkServiceConnection.isBound) {
            if (location != null) {
                val serverIp = preferenceService.getString(R.string.pkey_server_ip)
                val serverPort = preferenceService.getString(R.string.pkey_server_port)
                networkServiceConnection.service?.connect(serverIp, Integer.valueOf(serverPort))
            }
        } else {
            bindNetworkService()
        }
    }

    private fun handleStatusChanged(status: Int) {
        color_surface.drawColor(0, 0, 0)
        val message = when (status) {
            CwsNetworkService.CONNECTED -> {
                connected = true
                autoHide = true
                networkServiceConnection.service?.sendMessage("pos ${location?.latitude} ${location?.longitude}")
                hide()
                connection_status.visibility = View.GONE
                getString(R.string.status_connected)
            }
            CwsNetworkService.DISCONNECTED -> {
                connected = false
                autoHide = false
                show()
                connection_status.visibility = View.VISIBLE
                getString(R.string.status_disconnected)
            }
            else -> {
                connected = false
                autoHide = false
                show()
                connection_status.visibility = View.VISIBLE
                getString(R.string.status_connecting)
            }
        }
        updateConnectionStatus(message)
    }

    private fun handleData(data: String) {
        val tokens = data.split(",").toTypedArray()
        if (tokens[0] == "c") {
            if (!streaming) {
                runOnUiThread {
                    color_surface.layoutParams = FrameLayout.LayoutParams(-1, -1, Gravity.CENTER)
                }
                streaming = true
            }
            handleColor(tokens)
        } else if (tokens[0] == "s") {
            if (streaming) {
                runOnUiThread {
                    color_surface.layoutParams = FrameLayout.LayoutParams(300, 300, Gravity.CENTER)
                }
                streaming = false
            }
            handleSequence(tokens)
        } else if (tokens[0] == "pos") {
            handlePosition(tokens)
        }
    }

    private fun handlePosition(tokens: Array<String>) {
        if (tokens[1] == "0") {
            networkServiceConnection.service?.disconnect()
            updateConnectionStatus("Location rejected")
        }
    }

    private fun handleColor(tokens: Array<String>) {
        val r = Integer.valueOf(tokens[1])
        val g = Integer.valueOf(tokens[2])
        val b = Integer.valueOf(tokens[3])
        color_surface.drawColor(r, g, b)
    }

    private fun handleSequence(tokens: Array<String>) {
        val fps = Integer.valueOf(tokens[1])!!
        val sequence = tokens[2]
        DisplaySequenceTask(
                color_surface,
                (1000 / fps).toLong()
        ).executeOnExecutor(AsyncTask.THREAD_POOL_EXECUTOR, sequence)
    }

    override fun onCreateOptionsMenu(menu: Menu): Boolean {
        menuInflater.inflate(R.menu.main, menu)
        return true
    }

    override fun onOptionsItemSelected(item: MenuItem): Boolean {
        val id = item.itemId
        if (id == R.id.action_settings) {
            val intent = Intent(this, SettingsActivity::class.java)
            startActivity(intent)
            return true
        }
        return super.onOptionsItemSelected(item)
    }

    private fun updateConnectionStatus(text: String) {
        if (connectionStatusToast != null) connectionStatusToast!!.cancel()
        connectionStatusToast = Toast.makeText(this, text, Toast.LENGTH_LONG)
        connectionStatusToast!!.show()
        connection_status.text = text
    }

    private fun onServiceConnected(connection: ServiceConnection) {
        if (connection is NetworkServiceConnection) {
            connect()
        }
    }

    internal class NetworkServiceConnection(activity: StreamActivity) : ServiceConnection {

        private val activityRef = WeakReference(activity)

        internal var isBound: Boolean = false
        var service: CwsNetworkService? = null
        private var dataSubscription: Disposable? = null
        private var statusSubscription: Disposable? = null

        override fun onServiceConnected(
            className: ComponentName,
            binder: IBinder
        ) {
            service = (binder as CwsNetworkService.LocalBinder).service
            isBound = true
            if (activityRef.get() != null) {
                dataSubscription =
                        activityRef.get()!!.networkServiceConnection.service?.data?.subscribe(
                                activityRef.get()!!::handleData
                        )
                statusSubscription = service
                        ?.status
                        ?.observeOn(AndroidSchedulers.mainThread())
                        ?.subscribe(activityRef.get()!!::handleStatusChanged)
            }
            activityRef.get()?.onServiceConnected(this)
        }

        override fun onServiceDisconnected(name: ComponentName) {
            service = null
            isBound = false
            dataSubscription?.dispose()
            statusSubscription?.dispose()
        }
    }
}
