import logging
import sys
from functools import wraps
from typing import List

import cv2
import fire
import requests
from twisted.internet.selectreactor import SelectReactor

from crowdstream.services.camera import CameraService
from crowdstream.services.devices import DeviceService
from crowdstream.services.network import NetworkService

logging.basicConfig(stream=sys.stdout, level=logging.INFO)

logger = logging.getLogger(__name__)


class CrowdstreamServer:
    """
    Commands:

    Public:
    auth <password>: Authenticate as admin
    pos <gps coordinates>: Login as audience

    Authenticated:
    init: start initialization process
    stream <int> or <path>: stream from web cam or file
    stream disconnect: disconnect streaming

    """

    position_acceptance_radius = 0.001
    restrict_device_location = True
    auto_detect_location = True
    password = None
    longitude = 0
    latitude = 0

    def __init__(self):
        self._authenticated_clients = set()
        self._network_service = NetworkService(self.data_received)
        self._device_service = DeviceService(self._network_service, CameraService())
        self._running = False
        from twisted.internet import reactor
        self._reactor: SelectReactor = reactor

    def _detect_location(self):
        response = requests.get('https://api.ipdata.co').json()
        self.latitude = response['latitude']
        self.longitude = response['longitude']
        logger.info(f"detected location: {self.latitude}, {self.longitude}, ")

    def start(self, port: int = 8000, password: str = '123123'):
        logger.info('server started')
        if self.auto_detect_location:
            self._detect_location()
        self._running = True
        self.password = password
        self._network_service.start(port)

    def initialize(self, attempts):
        logger.debug('initialize')
        self._device_service.position_devices(attempts)

    def data_received(self, protocol, data: bytes):
        line = data.decode()
        tokens = line.split(' ')
        cmd = tokens[0]
        if cmd == 'auth':
            self._handle_auth(protocol, tokens)
        elif cmd == 'pos':
            self._handle_position(protocol, tokens)
        elif self._is_authenticated(protocol):
            self._handle_control_cmd(protocol, tokens)
        else:
            logger.info('Unauthorized access attempted')

    def _invalid_input(self, protocol, line):
        if self._is_authenticated(protocol):
            protocol.sendLine(f"invalid command: {line}".encode())

    def _handle_control_cmd(self, protocol, tokens: List[str]):
        cmd = tokens[0]
        if cmd == 'init':
            # self.initialize()
            self._initialization_thread(tokens)
            result = True
        elif cmd == 'stream':
            result = self._handle_stream(tokens)
        elif cmd == 'status':
            result = self.handle_status(protocol)
        else:
            logging.info('invalid command')
            result = False
        if result:
            protocol.sendLine('ok'.encode())
        else:
            self._invalid_input(protocol, tokens)

    def _handle_auth(self, protocol, tokens):
        if len(tokens) == 2 and tokens[1] == self.password:
            self._authenticate(protocol)
            protocol.sendLine('authentication successful'.encode())
            logger.info("authentication successful")
        else:
            protocol.sendLine("authentication failed".encode())
            logger.info("authentication failed")

    def _initialization_thread(self, tokens):
        attempts = 3
        if len(tokens) == 2:
            attempts = int(tokens[1])
        self._reactor.callInThread(self.initialize, attempts)

    def _stream_thread(self, source):
        logger.info(source)
        # noinspection PyArgumentList
        self._reactor.callInThread(self._device_service.project_video, cv2.VideoCapture(source))

    def _handle_stream(self, tokens) -> bool:
        if len(tokens) != 2:
            return False
        arg = tokens[1]
        if arg == 'stop':
            self._device_service.stream = False
        else:
            try:
                source = int(arg)
            except ValueError:
                import os
                source = os.path.abspath(os.path.join(os.path.dirname(__file__), '../..', 'content', arg))
            self._stream_thread(source=source)
        return True

    def _handle_position(self, protocol, tokens):
        if len(tokens) == 3 and self._position_accepted(tokens[1:]):
            logger.debug(f"Device accepted: {protocol.transport.client[0]}")
            self._device_service.accept_device(protocol)
        else:
            logger.debug(f"Device rejected: {protocol.transport.client[0]}")
            self._device_service.reject_device(protocol)

    def _position_accepted(self, position_data: List[float]):
        if not self.restrict_device_location:
            return True
        try:
            latitude = float(position_data[0])
            longitude = float(position_data[1])
        except (ValueError, IndexError):
            return False
        return (longitude - self.longitude) ** 2 + (latitude - self.latitude) ** 2 < self.position_acceptance_radius

    def _authenticate(self, protocol):
        self._authenticated_clients.add(protocol)
        connection_lost = protocol.connectionLost

        @wraps(connection_lost)
        def _deauth(*args, **kwargs):
            result = connection_lost(*args, **kwargs)
            self._authenticated_clients.remove(protocol)
            return result

        protocol.connectionLost = _deauth

    def _is_authenticated(self, protocol):
        # TODO: remove auth when device disconnects
        return protocol in self._authenticated_clients

    def handle_status(self, protocol):
        protocol.sendLine(f"""total connections: {len(self._network_service.get_ips())}
authenticated clients: {len(self._authenticated_clients)}
connected devices: {len(self._device_service.get_active_ips())}
mapped devices: {len(self._device_service.devices)}"""
                          .encode())
        return True


def start(port: int = 8000, password: str = '123123'):
    server = CrowdstreamServer()
    server.start(port, password)


def main():
    fire.Fire(start)


if __name__ == "__main__":
    main()
