import logging
from threading import Condition
from typing import List, Dict

import cv2
import numpy as np

from crowdstream import config
from crowdstream.config import EXTEND_CAPTURE_SECONDS
from crowdstream.services.camera import CameraService
from crowdstream.services.network import NetworkService
from crowdstream.utils.debug import debug_tracking
from crowdstream.utils.sequences import generate_sequences, encode_and_assign_sequences
from crowdstream.utils.tracking import Track
from crowdstream.utils.types import Color, Point
from crowdstream.utils.video import write_video

logger = logging.getLogger(__name__)


class Device:
    _sample_rate = config.SYMBOL_FREQUENCY
    _sample_interval = 1 / _sample_rate
    _source_fps = 30
    _sample_frame = int(_source_fps / _sample_rate)

    def __init__(self, ip: str, position: 'Point'):
        self._shape = config.CAMERA_DIMENSIONS
        self._source_position = None
        self._source_shape = None
        self.transposed_start: 'Point' = None
        self.transposed_end: 'Point' = None
        self.ip = ip
        self.position = position

    def transpose(self, shape):
        width, height = self._shape
        width_factor = shape[1] / width
        height_factor = shape[0] / height
        (x, y) = self.position
        input_x = int(x * width_factor)
        input_y = int(y * height_factor)
        self._source_position = (input_x, input_y)
        start_y = max(input_y - config.SAMPLE_RECTANGLE_SIZE, 0)
        start_x = max(input_x - config.SAMPLE_RECTANGLE_SIZE, 0)
        self.transposed_start = Point(start_x, start_y)
        end_y = min(input_y + config.SAMPLE_RECTANGLE_SIZE, height)
        end_x = min(input_x + config.SAMPLE_RECTANGLE_SIZE, width)
        self.transposed_end = Point(end_x, end_y)


class DeviceService:
    DEBUG = False
    DEBUG_FRAME_DELAY = 10
    filename = r"../input"

    def __init__(self, network_service: NetworkService, camera_service: CameraService):
        """
        :param network_service:
        :param camera_service:
        """
        self.default_warm_up_frames = config.DEFAULT_WARMUP_FRAMES
        self.devices: List[Device] = []
        self._network_service: NetworkService = network_service
        self._camera_service: CameraService = camera_service
        self._shape: (int, int, int) = None
        self._video = False
        self._stream = False
        self._sequence_length = config.SEQUENCE_LENGTH
        self._sequence_prefix = config.SEQUENCE_PREFIX
        self._timestamps = []
        self._recording = False
        self._cv = Condition()
        self._ips = {}

    def position_devices(self, attempts=config.DEFAULT_MAX_INIT_ROUNDS):
        self._video = False
        self._stream = False
        self.devices: List[Device] = []
        self._shape: (int, int, int) = None
        for _ in range(attempts):
            # TODO: use IPs from Devices in self.devices instead
            # _ips = self.network_service.get_ips()
            ips = set(self.get_active_ips())
            # TODO: keep track of mapped IPs in some other way, another map or check Device.pos/track
            mapped_ips = set([device.ip for device in self.devices])
            ips -= mapped_ips
            self.devices.extend(self._locate_devices(ips))
        logger.info("total: mapped %d out of %d devices", len(self.devices), len(self.get_active_ips()))

    def _generate_sequences(self, number, length, prefix, streak):
        return generate_sequences(number, length, prefix, streak)

    def _locate_devices(self, ips) -> List[Device]:
        """
        Run the device localization sequence
        :return:
        """
        if not ips:
            return []
        logger.debug("_locate devices")
        sequences = self._generate_sequences(len(ips), self._sequence_length, self._sequence_prefix,
                                             config.MAX_SYMBOL_STREAK)
        sequence_map, encoded_sequence_map = encode_and_assign_sequences(sequences, ips)
        success = self._send_and_capture_sequence(self._camera_service.get_camera(), encoded_sequence_map)
        if not success:
            logger.error("Unable to record sequence")
            return ips
        # noinspection PyArgumentList
        tracks = self._track_devices(cv2.VideoCapture(f"{self.filename}.mp4"), self._timestamps)
        devices = self._map_devices(sequence_map, tracks)
        logger.info("round: mapped %d out of %d devices", len(devices), len(ips))
        return devices

    def _send_and_capture_sequence(self, cap, sequence_map: Dict[str, str]):
        logger.info("_send_and_capture_sequence")
        duration = 1 / config.SYMBOL_FREQUENCY * config.SEQUENCE_LENGTH + EXTEND_CAPTURE_SECONDS
        from threading import Thread
        recording_thread = Thread(target=self._record, args=(cap, duration, self.filename))
        recording_thread.start()
        with self._cv:
            success = self._cv.wait_for(lambda: self._recording, duration)
        logger.info('recording: ' + str(success))
        self._network_service.send_sequences(symbol_frequency=config.SYMBOL_FREQUENCY, sequence_map=sequence_map)
        recording_thread.join(duration * 2)
        return not recording_thread.is_alive()

    def _track_devices(self, cap: cv2.VideoCapture, timestamps) -> List[Track]:
        logger.info("_track_devices")
        frames = []
        detections = []
        while cap.isOpened():
            success, frame = cap.read()
            if not success:
                break
            if self.DEBUG:
                frames.append(frame)
            from crowdstream.utils import detection
            detections.append(detection.detect_devices(frame, config.DETECTION_THRESHOLD))

        cap.release()
        from crowdstream.utils import tracking
        tracks = tracking.perform_tracking(detections, timestamps)

        if self.DEBUG:
            debug_tracking(frames, timestamps, tracks, self.DEBUG_FRAME_DELAY)
        return tracks

    def stop(self):
        self._stream = False

    def project_image(self, source: np.ndarray):
        if self._shape != source.shape:
            for device in self.devices:
                device.transpose(source.shape)
                self._shape = source.shape
        color_map: Dict[Device, (int, int, int)] = {}
        for device in self.devices:
            # TODO: extract single point
            region = source[device.transposed_start.y:device.transposed_end.y,
                     device.transposed_start.x:device.transposed_end.x]
            color = tuple(int(scalar) for scalar in cv2.mean(region))
            color_map[device] = color
        self._send_colors_to_devices(color_map)

    def _send_colors_to_devices(self, color_map: Dict[Device, Color]):
        for device, color in color_map.items():
            self._network_service.send_color_to_ip(color, device.ip)

    def project_video(self, cap: cv2.VideoCapture):
        logger.debug("project_video")
        self._video = True
        self._stream = True
        fps = cap.get(cv2.CAP_PROP_FPS)
        fps = fps if fps > 10 else 10
        seconds_between_frames = 1 / fps
        while self._stream and cap.isOpened():
            import time
            time.sleep(seconds_between_frames)
            success, frame = cap.read()
            if not success:
                continue
            self.project_image(frame)
        self._video = False
        color_map = {}
        for device in self.devices:
            color_map[device] = (0, 0, 0, 0)
        self._send_colors_to_devices(color_map)
        logger.debug("stop project_video")

    def _map_devices(self, ip_map: Dict[str, str], tracks: List[Track]) -> List[Device]:
        """
        Maps tracks to IPs and creates a Device for each successful mapping
        :param ip_map: A mapping from sequences to IPs
        :param tracks: A list of tracks
        :return: The devices
        """
        logger.debug("_map_devices")
        num_matched = 0
        devices = []
        for track in tracks:
            sequence = track.get_sequence(self._sequence_length)
            ip = ip_map.pop(sequence, None)
            if ip:
                devices.append(Device(ip, track.position))
                num_matched += 1

            else:
                logger.debug(f"no match:(['{sequence}")
                logger.debug(str(ip_map.keys()))
        logger.info("mapped %d devices", num_matched)
        if len(devices):
            logger.info(f"mapped {num_matched} of {len(tracks)} devices ({num_matched/len(tracks)*100}%)")
        return devices

    def _record(self, cap: cv2.VideoCapture, duration, filename: str):
        """
        Record frames from send capture
        :param cap: the capture device
        :return:
        """
        logger.debug("_record")
        if not cap.isOpened():
            logger.error("Could not open camera for recording")
            cap.release()
            return
        width = int(cap.get(cv2.CAP_PROP_FRAME_WIDTH))
        height = int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
        size = width, height

        self._timestamps = []
        import time
        start = None
        timestamp = 0
        end = None
        capture_fps = cap.get(cv2.CAP_PROP_FPS)
        static_time_between_frames = 1 / capture_fps if capture_fps else 0
        frames = []
        warm_up_frames = self.default_warm_up_frames
        while (not end or timestamp < end) and cap.isOpened():
            success, frame = cap.read()
            if not success:
                break
            if warm_up_frames:
                warm_up_frames -= 1
                continue
            timestamp = timestamp + static_time_between_frames if static_time_between_frames else time.time()
            if start is None:
                start = timestamp
                end = start + duration if duration else None
                with self._cv:
                    self._recording = True
                    self._cv.notify()
            self._timestamps.append(timestamp - start)
            frames.append(frame)
        cap.release()
        self._recording = False
        time_spent = timestamp - start
        fps = round(len(frames) / time_spent)
        # noinspection PyArgumentList
        write_video(fps, frames, size, filename)
        with open(f"{filename}-timestamps.txt", 'w') as f:
            import json
            f.write(json.dumps(self._timestamps))

        logger.info(f"Recorded {len(frames)} frames in {time_spent} seconds, {fps} fps")

    def get_active_ips(self) -> List[str]:
        return list(self._ips.keys())

    def accept_device(self, protocol):
        self._ips[protocol.transport.client[0]] = protocol

    @staticmethod
    def reject_device(protocol):
        protocol.sendLine("pos 0".encode())
        protocol.transport.abortConnection()
