import logging
import struct
from typing import Tuple, Dict, Set

from twisted.internet import interfaces, reactor
from twisted.internet.endpoints import TCP4ServerEndpoint
from twisted.internet.protocol import Factory, Protocol, connectionDone
# noinspection PyRedeclaration
from twisted.protocols.basic import LineOnlyReceiver


class CrowdstreamProtocol(LineOnlyReceiver):
    def lineReceived(self, line):
        if self.callback:
            self.callback(self, line)

    def __init__(self, data_received_cb, connection_made_cb, connection_lost_cb):
        self.connection_made_cb = connection_made_cb
        self.connection_lost_cb = connection_lost_cb
        self.callback = data_received_cb

    def connectionMade(self):
        self.connection_made_cb(self)

    def connectionLost(self, reason=connectionDone):
        self.connection_lost_cb(self)


class NetworkService(Factory):
    logger = logging.getLogger(__name__)

    def __init__(self, callback):
        self._connections: Dict[str, CrowdstreamProtocol] = {}
        self._callback = callback
        self._rgb_struct = struct.Struct('3B')
        self._buffer = bytearray(self._rgb_struct.size)
        # noinspection PyProtectedMember
        from twisted.internet.selectreactor import SelectReactor
        self.reactor: SelectReactor = reactor

    def get_ips(self) -> Set[str]:
        """
        Get a list of all connected IPs
        :return: A list of all connected IPs
        """
        return set(self._connections.keys())

    def send_color_to_ip(self, color: Tuple[int, int, int, int], ip: str):
        protocol = self._connections.get(ip, None)
        if protocol:
            b, g, r, a = color
            data = f"c,{r},{g},{b}"
            self.reactor.callFromThread(self._write_string_to_protocol, protocol, data)

    @staticmethod
    def _write_to_protocol(protocol: CrowdstreamProtocol, data):
        protocol.sendLine(data.en)

    @staticmethod
    def _write_string_to_protocol(protocol: CrowdstreamProtocol, msg: str):
        protocol.sendLine(msg.encode())

    def send_sequences(self, symbol_frequency: int, sequence_map: Dict[str, str]):
        for sequence, ip in sequence_map.items():
            protocol = self._connections[ip]
            if protocol:
                data = f"s,{symbol_frequency},{sequence}"
                self.reactor.callFromThread(self._write_string_to_protocol, protocol, data)

    def start(self, port):
        endpoint = TCP4ServerEndpoint(self.reactor, port)
        endpoint.listen(self)
        self.reactor.suggestThreadPoolSize(30)
        self.reactor.run()

    def buildProtocol(self, address: interfaces.IAddress):
        protocol = CrowdstreamProtocol(self._callback, self.connection_made, self.connection_lost)
        return protocol

    def connection_made(self, protocol: Protocol):
        self.logger.info("connected: %s", protocol.transport.client[0])
        self._connections[protocol.transport.client[0]] = protocol

    def connection_lost(self, protocol: Protocol):
        # TODO: handle same IP connections
        self.logger.info("disconnected: %s", protocol.transport.client[0])
        self._connections.pop(protocol.transport.client[0], None)
