import functools
from typing import List

import cv2
import numpy as np

from crowdstream.utils import draw_circle
from crowdstream.utils.types import Point

DEBUG_DETECTION = False
DEBUG_THRESHOLD = False
DEBUG_CONTOURS = False

DEFAULT_THRESHOLD = 200

Contours = List[List[Point]]


def draw_debug(frame, position):
    draw_circle(frame, position, (0, 255, 0))


DEBUG_FUNCTION = draw_debug


def detect_devices2(input_image, threshold_func, find_contours_func, contour_center_func) -> List[Point]:
    img = cv2.cvtColor(input_image, cv2.COLOR_BGR2GRAY)
    img = threshold_func(img)
    contours = find_contours_func(img)
    positions = contour_center_func(contours)
    if DEBUG_DETECTION:
        for position in positions:
            DEBUG_FUNCTION(frame=input_image, position=position)
        cv2.imshow("detection", input_image)
    return positions


def detect_devices(input_image: np.ndarray, threshold: int = DEFAULT_THRESHOLD) -> List[Point]:
    return detect_devices2(input_image, functools.partial(simple_threshold, threshold=threshold),
                           external_approx_contours, minimum_enclosing_circle)


def simple_threshold(image, threshold) -> np.ndarray:
    _, output = cv2.threshold(image, threshold, 255, cv2.THRESH_BINARY)  # type: (int, np.ndarray)
    if DEBUG_THRESHOLD:
        cv2.imshow('threshold', output)
    return output


def otsu_threshold(image) -> np.ndarray:
    _, output = cv2.threshold(image, 0, 255, cv2.THRESH_BINARY + cv2.THRESH_OTSU)  # type: (int, np.ndarray)
    if DEBUG_THRESHOLD:
        cv2.imshow('Otsu`s threshold', output)
    return output


def gaussian_otsu_threshold(image) -> np.ndarray:
    blur = cv2.GaussianBlur(image, (5, 5), 0)
    return otsu_threshold(blur)


def external_approx_contours(image) -> Contours:
    contours_img, contours, _ = cv2.findContours(image, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
    if DEBUG_CONTOURS:
        cv2.imshow('contours', contours_img)
    return contours


def external_contours(image) -> Contours:
    contours_img, contours, _ = cv2.findContours(image, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_NONE)
    if DEBUG_CONTOURS:
        cv2.imshow('contours', contours_img)
    return contours


def minimum_enclosing_circle(contours: Contours) -> List[Point]:
    centers = []
    for cnt in contours:
        (x, y), _ = cv2.minEnclosingCircle(cnt)
        centers.append((int(x), int(y)))
    return centers


def contour_moments_centroid(contours) -> List[Point]:
    centres = []
    for contour in contours:
        moments = cv2.moments(contour, True)
        if moments['m00'] != 0:
            centres.append((int(moments['m10'] / moments['m00']), int(moments['m01'] / moments['m00'])))
    return centres
