import logging
from collections import namedtuple
from typing import Dict, List, Tuple, Union

from scipy.spatial.ckdtree import cKDTree

from crowdstream import config
from crowdstream.config import TEST_DEFAULT_VIDEO_FPS
from crowdstream.utils.types import Point

COLOR_TRACK_FOUND = (0, 255, 0)

COLOR_TRACK_LOST = (0, 0, 255)
State: Tuple[bool, Point, float] = namedtuple('State', ['visible', 'position', 'timestamp', 'state'])

logger = logging.getLogger(__name__)
logger.setLevel(logging.WARN)

DEBUG_SHOW_SEARCH_RADIUS = True
# Search distance increase per second since last frame, reasonable values are 80-200 depending on camera distance
SEARCH_RADIUS_EXPANSION_RATE = 200

MAX_SEARCH_RADIUS = 20.0

MAX_HIDDEN_TIME = 1 / config.SYMBOL_FREQUENCY * (config.MAX_SYMBOL_STREAK + 1)


class Track:
    sample_rate = config.SYMBOL_FREQUENCY
    source_fps = config.TEST_DEFAULT_VIDEO_FPS
    _sample_interval = 1.0 / sample_rate
    _sample_frame = int(source_fps / sample_rate)

    def __init__(self, position: Point):
        self.predicted_position = None
        self.predicted_symbol = None
        self._new_position = position
        self.sequence = []
        self.track = []
        self.last_seen = 0
        self.visible = True
        self._next_sample_time = None
        self.frame_counter = 0
        self.position = None
        self.history: List[State] = []
        self.last_sample_index = 0
        self.samples = []

    def predict_state(self):
        self.predicted_symbol = self.sequence[-1]

    def get_visible_probability(self):
        pass

    def get_predicted_position(self):
        self.predicted_position = self.track[-1]

    def queue_update(self, new_position: Union[Point, None]):
        self._new_position = new_position

    def apply_update(self, timestamp: float):
        if self._new_position:
            self.position = Point(x=self._new_position[0], y=self._new_position[1])
            self._new_position = None
            self.track.append(self.position)
            self.visible = True
            self.last_seen = timestamp
        else:
            self.visible = False
        if not self._next_sample_time:
            self._next_sample_time = timestamp + self._sample_interval
        elif timestamp >= self._next_sample_time:
            last_states = self.history[self.last_sample_index:]
            num_visible_samples = 0
            num_hidden_samples = 0
            closest = (last_states[0].timestamp, last_states[0].visible)
            center_unit_frame = self._next_sample_time - (self._sample_interval / 2)
            for state in last_states:
                if abs(state.timestamp - center_unit_frame) < abs(closest[0] - center_unit_frame):
                    closest = (state.timestamp, state.visible)
                if state.visible:
                    num_visible_samples += 1
                else:
                    num_hidden_samples += 1
            if num_hidden_samples == num_visible_samples:
                logger.debug(f"Equal number of hidden({num_hidden_samples}) and visible({num_visible_samples}) states")
                self.sequence.append('1' if closest[1] is True else '0')
            else:
                self.sequence.append('1' if num_visible_samples > num_hidden_samples else '0')
            self._next_sample_time += self._sample_interval
            self.last_sample_index = self.frame_counter
        self.frame_counter += 1
        last_state = self.sequence[-1] if self.sequence else False
        self.history.append(State(self.visible, self.position, timestamp, last_state))

    def get_sequence(self, length=None):
        return ''.join(str(sample) for sample in self.sequence)[:length]


def perform_tracking(frame_detections: List[List[Point]],
                     timestamps: List[float]
                     ) -> List[Track]:
    tracks = []
    for index, detections in enumerate(frame_detections):
        _update(detections, tracks, timestamps[index])
        for track in tracks:
            track.apply_update(timestamps[index])
    return tracks


def _update(detected_positions: List[Point],
            tracks: List[Track],
            timestamp: float):
    # TODO: utilize information we know
    # Only 3 of the same symbol in a row, track can only be visible or hidden for 3 symbols
    hidden_devices: List[Track] = []
    used_positions: Dict[int, Tuple[Track, float, int]] = {}
    remaining_positions = detected_positions.copy()
    tree = cKDTree(detected_positions, leafsize=16, compact_nodes=True,
                   balanced_tree=True, copy_data=False) if detected_positions else None
    for track in tracks:
        if config.MAP_VISIBLE_FIRST and not track.visible:
            hidden_devices.append(track)
            continue

        if not detected_positions:
            track.queue_update(None)
            continue

        _associate(detected_positions, remaining_positions, timestamp, track, tree, used_positions, 1)

    new_device_positions = _track_hidden(hidden_devices, remaining_positions, timestamp)
    for position in new_device_positions:
        tracks.append(Track(position))


def _associate(detected_positions: List[Point], remaining_positions: List[Point], timestamp: float, track: Track,
               tree: cKDTree, used_positions: Dict, neighbor_index: int):
    time_since_last_detection = timestamp - track.last_seen
    if time_since_last_detection > MAX_HIDDEN_TIME:
        logger.debug("Skipping permanently lost track")
        return

    search_distance = min(SEARCH_RADIUS_EXPANSION_RATE * time_since_last_detection, MAX_SEARCH_RADIUS)
    logger.debug("Search distance: %d", search_distance)
    dd, ii = tree.query(track.position, k=[neighbor_index],
                        distance_upper_bound=search_distance)  # type: (List[float], List[int])
    distance = dd[0]
    index = ii[0]
    if distance > search_distance:
        track.queue_update(None)
        return
    position = detected_positions[index]
    if used_positions.get(index, None):
        other_device, other_distance, other_neighbour, = used_positions.get(index)
        this_next_neighbor = tree.query(track.position, k=[neighbor_index + 1])[0]  # type: (float, int)
        other_next_neighbor = tree.query(other_device.position, k=[other_neighbour + 1])[0]  # type: (float, int)

        # TODO: tests different ways of selecting which track maps to which point, eg. min distance to closest, min average distance
        # this_is_closer = this_next_neighbor[0] + other_distance > distance + other_next_neighbor[0]
        this_is_closer = distance < other_distance
        if this_is_closer:
            used_positions[index] = (track, distance, neighbor_index)
            track.queue_update(position)
            _associate(detected_positions, remaining_positions, timestamp, other_device, tree, used_positions,
                       neighbor_index=other_neighbour + 1)
        else:
            _associate(detected_positions, remaining_positions, timestamp, track, tree, used_positions,
                       neighbor_index=neighbor_index + 1)
    else:
        used_positions[index] = (track, distance, neighbor_index)
        track.queue_update(position)
        remaining_positions.remove(position)


def _track_hidden(hidden_tracks: List[Track], detected_positions: List[Point], timestamp: float):
    new_device_positions = detected_positions.copy()
    if any(hidden_tracks) and any(detected_positions):
        tree = cKDTree(detected_positions)
        used_positions: Dict[int, Track] = {}
        for track in hidden_tracks:
            _associate(detected_positions, new_device_positions, timestamp, track, tree, used_positions, 1)
    return new_device_positions
