import functools
from typing import List, Tuple
from unittest import TestCase

import cv2
import matplotlib.pyplot as plt
import pytest
from scipy.spatial.ckdtree import cKDTree

from crowdstream.utils import detection, debug
from tests.test_utils import test_generator, TESTDATA_DIR


def _print_detection_report(points):
    print(f"\nDetected devices: {len(points)}")


def _show_video_detection_report(detections: List[Tuple[int, int]]):
    # TODO: implement this
    pass


@pytest.fixture(scope="class", params=['1.jpg', '2.jpg', '3.jpg'])
def file(request):
    test_image_path = rf'..\..\testdata\image\{request.param}'
    return cv2.imread(test_image_path)


@pytest.fixture(scope="class",
                params=[functools.partial(detection.simple_threshold, threshold=250), detection.otsu_threshold,
                        detection.gaussian_otsu_threshold])
def threshold_func(request):
    return request.param


@pytest.fixture()
def find_contours_func():
    return detection.external_approx_contours


@pytest.fixture()
def find_contour_centre_func():
    return detection.minimum_enclosing_circle


@pytest.fixture()
def detection_func(threshold_func, find_contours_func, find_contour_centre_func):
    return functools.partial(detection.detect_devices2, threshold_func=threshold_func,
                             find_contours_func=find_contours_func, contour_center_func=find_contour_centre_func)


def test__(detection_func, file):
    points = detection_func(input_image=file)
    print(detection_func)
    _print_detection_report(points)
    cv2.waitKey(0)


def test__generated(detection_func):
    timestamps = test_generator.generate_video((1, 1), '../../testdata/detection/')


def test__process_sequencing__real_video():
    cap: cv2.VideoCapture = cv2.VideoCapture(rf"{TESTDATA_DIR}/1.mp4")
    detections = []
    frames = []
    while cap.isOpened():
        success, frame = cap.read()
        if not success:
            break
        frames.append(frame)
        detections.append(detection.detect_devices(frame, 180))
    debug.debug_detections(frames, detections)


class ManualThresholdTest(TestCase):

    @classmethod
    def setUpClass(cls):
        cls.threshold_func = functools.partial(detection.simple_threshold, threshold=250)
        cls.contours_func = detection.external_approx_contours
        cls.contour_centre_func = detection.minimum_enclosing_circle


class OtsuTest(TestCase):

    def setUp(self):
        super().setUp()


def test_detection_real_image_1():
    detection.DEBUG_DETECTION = True
    detection.DEBUG_CONTOURS = True
    detection.DEBUG_THRESHOLD = True
    test_image_path = r'..\testdata\image\1.jpg'
    img = cv2.imread(test_image_path)
    points = detection.detect_devices(img)
    _print_detection_report(points)
    cv2.waitKey(0)


def test_detection_real_image_2():
    detection.DEBUG_DETECTION = True
    detection.DEBUG_CONTOURS = True
    detection.DEBUG_THRESHOLD = True
    test_image_path = r'..\testdata\image\2.jpg'
    img = cv2.imread(test_image_path)
    points = detection.detect_devices(img)
    _print_detection_report(points)
    cv2.waitKey(0)


def test_detection_real_image_3():
    detection.DEBUG_DETECTION = True
    detection.DEBUG_CONTOURS = True
    detection.DEBUG_THRESHOLD = True
    test_image_path = r'..\testdata\image\3.jpg'
    img = cv2.imread(test_image_path)
    points = detection.detect_devices(img)
    _print_detection_report(points)
    cv2.waitKey(0)


def test_detection_real_image_4():
    detection.DEBUG_DETECTION = True
    detection.DEBUG_CONTOURS = False
    detection.DEBUG_THRESHOLD = True
    test_image_path = r'..\..\testdata\image\4.jpg'
    img = cv2.imread(test_image_path)
    points = detection.detect_devices(img)
    _print_detection_report(points)
    cv2.waitKey(0)


def test_detection_generated_10x10():
    # noinspection PyArgumentList
    cap = cv2.VideoCapture(rf"..\testdata\video\test_detection_generated_10x10.avi")
    frames = []
    while cap.isOpened:
        success, frame = cap.read()
        if not success:
            break
        devices = detection.detect_devices(frame)
        frames.append(devices)
        assert len(devices) == 100
    _show_video_detection_report(frames)


def test_detection_generated_20x20():
    detection.DEBUG_DETECTION = True
    # noinspection PyArgumentList
    cap = cv2.VideoCapture(rf"..\testdata\video\test_detection_generated_20x20.avi")
    frames = []
    while cap.isOpened:
        success, frame = cap.read()
        if not success:
            break
        detections = detection.detect_devices(frame)
        frames.append(detections)
        if len(detections) < 20 * 20:
            cv2.waitKey(0)
    _show_video_detection_report(frames)


def test_detection_live():
    detection.DEBUG_DETECTION = True
    detection.THRESHOLD = 220
    # noinspection PyArgumentList
    cap = cv2.VideoCapture(1)
    while cap.isOpened:
        success, frame = cap.read()
        if not success:
            break
        cv2.imshow('test', frame)
        cv2.waitKey(10)


def test__detection_precision():
    offset = []
    movements = []
    for movement in range(0, 50, 3):
        filename = rf'{TESTDATA_DIR}/tracking/generated'
        timestamps, positions, sequences, actual_tracks = test_generator.generate_video((10, 10), filename, movement=5,
                                                                                        max_movement_distance=30,
                                                                                        device_size_range=(movement, movement))
        cap = cv2.VideoCapture(f"{filename}.mp4")
        detections = []
        while cap.isOpened():
            success, frame = cap.read()
            if not success:
                break
            detections.append(detection.detect_devices(frame))
        total = 0
        count = 0
        for i, pos in enumerate(positions):
            if not pos:
                continue
            count += 1
            tree = cKDTree(pos)
            total_distance = 0
            for position in detections[i]:
                dd, ii = tree.query(position)
                total_distance += dd
            # if total_distance != 0:
            #     print(f"{total_distance/len(pos)}")

            total += (total_distance / len(pos))
        movements.append(movement)
        offset.append(total / count)

        # print(f"Tracking: {total/count}")

    plt.plot(movements, offset)
    plt.ylabel("precision error")
    plt.xlabel("size")
    plt.show()
