import math
import random
from typing import List, Dict, Tuple

import cv2
import numpy as np
from scipy.spatial.distance import cdist

from crowdstream import config
from crowdstream.utils.sequences import generate_sequences
from crowdstream.utils.video import create_empty_frame
from tests.test_utils import file


def _get_point_in_radius(position: Tuple[int, int], radius) -> Tuple[int, int]:
    if radius < 1:
        return position
    u = np.random.random() + np.random.random()
    r = radius * (2 - u if u > 1 else u)
    alpha = 2 * math.pi * np.random.random()
    x = r * math.cos(alpha) + position[0]
    y = r * math.sin(alpha) + position[1]
    return round(x), round(y)


def _get_point_color(random_part: int) -> Tuple[int, int, int]:
    rnd_intensity = (int(random.random() * random_part),
                     int(random.random() * random_part),
                     int(random.random() * random_part))
    base_intensity = 255 - random_part
    return base_intensity + rnd_intensity[0], base_intensity + rnd_intensity[1], base_intensity + rnd_intensity[2]


def _get_point_draw_radius(radius: Tuple[int, int]) -> int:
    min_radius, max_radius = radius
    return int(min_radius + (
            max_radius - min_radius) * random.random())


def _add_random_delay(sequence, max_delay):
    padding = random.randrange(0, max_delay) if max_delay else 0
    padding_after = max_delay - padding + 5
    return ("0" * padding) + sequence + ("0" * padding_after)


def generate_video(devices_shape: Tuple[int, int], path: str, sequences: List[str] = None, fps: int = 30,
                   movement: int = 2,
                   symbol_frequency: int = config.SYMBOL_FREQUENCY, device_color_variation: int = 50,
                   device_size_range: Tuple[int, int] = (4, 10),
                   output_shape: Tuple[int, int] = (1080, 720), max_movement_distance: int = 20, max_start_delay=10):
    sequences = sequences if sequences else generate_sequences(devices_shape[0] * devices_shape[1])

    modified_sequences = [_add_random_delay(sequence, max_start_delay) for sequence in sequences]
    (width, height) = output_shape
    (devices_x, devices_y) = devices_shape
    step_x = round(width / devices_x)
    step_y = round(height / devices_y)
    sequence_map = {}  # type: Dict[int, str]
    frames_per_symbol = int(fps / symbol_frequency)
    total_frames = frames_per_symbol * len(modified_sequences[0])
    last_position = {}  # type: Dict[int, Tuple[int, int]]
    initial_position = {}  # type: Dict[int, Tuple[int, int]]
    timestamps = []
    timestamp = 0
    all_positions = []
    frames = []
    ground_truths = {'frames': frames, 'class': 'video'}

    timestamp_difference = 1 / fps
    # noinspection PyArgumentList
    writer = cv2.VideoWriter(f"{path}.mp4", cv2.VideoWriter_fourcc(*"mp4v"), fps, (width, height), True)
    for frame_count in range(total_frames):
        timestamps.append(timestamp)
        device_index = 0
        frame_data = create_empty_frame(width, height)
        frame_positions = []
        annotations = []
        frame = {'timestamp': timestamp, 'num': frame_count, 'class': 'frame', 'annotations': annotations}
        timestamp += timestamp_difference

        for x in range(int(step_x / 2), width, step_x):
            for y in range(int(step_y / 2), height, step_y):
                if frame_count == 0:
                    sequence_map[device_index] = modified_sequences[device_index]  # assign sequences
                    last_position[device_index] = _get_point_in_radius((x, y), 0)
                    initial_position[device_index] = last_position[device_index]
                dist_from_origin = None
                new_position = None
                while dist_from_origin is None or dist_from_origin > max_movement_distance:
                    new_position = _get_point_in_radius(last_position[device_index], movement)
                    dist_from_origin = cdist(np.array([initial_position[device_index]]), np.array([new_position]))[0]
                annotation = {'height': 1, 'width': 1, 'id': device_index, 'dco': True, 'x': last_position[device_index][0], 'y': last_position[device_index][1]}
                symbol = sequence_map.get(device_index)[int(frame_count / frames_per_symbol)]
                if symbol == '1':
                    last_position[device_index] = new_position
                    # cv2.circle(frame, new_position, _get_point_draw_radius(device_size_range),
                    #            _get_point_color(device_color_variation), -1)
                    size = _get_point_draw_radius(device_size_range)
                    pt1 = (math.floor(new_position[0] - size / 2), math.floor(new_position[1] - size / 2))
                    pt2 = (math.ceil(new_position[0] + size / 2), math.ceil(new_position[1] + size / 2))
                    cv2.rectangle(frame_data, pt1, pt2, color=_get_point_color(device_color_variation),
                                  thickness=cv2.FILLED)
                    frame_positions.append(new_position)
                    # ground_truths[device_index].append(new_position)
                    annotation['x'] = new_position[0]
                    annotation['y'] = new_position[1]
                    annotation['dco'] = False
                # else:
                #     if ground_truths[device_index]:
                #         ground_truths[device_index].append(ground_truths[device_index][-1])
                    annotations.append(annotation)
                device_index += 1
        writer.write(frame_data)
        all_positions.append(frame_positions)
        frames.append(frame)
    writer.release()
    file.write_obj(f"{path}-timestamps.txt", timestamps)
    file.write_obj(f"{path}-positions.txt", all_positions)
    file.write_obj(f"{path}-sequences.txt", sequences)
    file.write_obj(f"{path}-groundtruth.txt", ground_truths)
    return timestamps, all_positions, sequences, ground_truths
