import itertools

import cv2

from crowdstream.services.devices import DeviceService
from crowdstream.utils import tracking, detection, debug
from pymot import MOTEvaluation
from tests.test_utils import TESTDATA_DIR, file, test_generator

DeviceService.DEBUG = True


def timestamp_generator(fps):
    index = 0.0
    while True:
        yield index * (1 / fps)
        index = index + 1


def test__process_sequencing__real_video():
    tracking.DEBUG = True
    cap: cv2.VideoCapture = cv2.VideoCapture(rf"{TESTDATA_DIR}/tracking/1.mp4")
    detections = []
    frames = []
    while cap.isOpened():
        success, frame = cap.read()
        if not success:
            break
        frames.append(frame)
        detections.append(detection.detect_devices(frame, 190))
    timestamps = list(itertools.islice(timestamp_generator(25), len(detections)))
    file.write_obj(rf"{TESTDATA_DIR}/tracking/1-positions.txt", detections)
    file.write_obj(rf"{TESTDATA_DIR}/tracking/1-timestamps.txt", timestamps)
    tracks = tracking.perform_tracking(detections, timestamps)
    debug.debug_tracking(frames, timestamps, tracks)


def test__process_sequencing__real_video_1():
    tracking.DEBUG = True
    cap: cv2.VideoCapture = cv2.VideoCapture(rf"{TESTDATA_DIR}/real1.avi")
    detections = []
    frames = []
    while cap.isOpened():
        success, frame = cap.read()
        if not success:
            break
        frames.append(frame)
        detections.append(detection.detect_devices(frame, 220))
    timestamps = file.read_obj(rf"{TESTDATA_DIR}/real1-timestamps.txt")
    sequences = file.read_obj(rf"{TESTDATA_DIR}/real1-sequences.txt")
    tracks = tracking.perform_tracking(detections, timestamps)
    debug.debug_tracking(frames, timestamps, tracks)
    _print_results(sequences, tracks)


def test__process_sequencing__real_video_2():
    tracking.DEBUG = True
    cap: cv2.VideoCapture = cv2.VideoCapture(rf"{TESTDATA_DIR}/real2.avi")
    detections = []
    frames = []
    while cap.isOpened():
        success, frame = cap.read()
        if not success:
            break
        frames.append(frame)
        detections.append(detection.detect_devices(frame, 220))
    timestamps = file.read_obj(rf"{TESTDATA_DIR}/real2-timestamps.txt")
    sequences = file.read_obj(rf"{TESTDATA_DIR}/real2-sequences.txt")
    tracks = tracking.perform_tracking(detections, timestamps)
    debug.debug_tracking(frames, timestamps, tracks)
    _print_results(sequences, tracks)


def test__process_sequencing__real_video_3():
    tracking.DEBUG = True
    cap: cv2.VideoCapture = cv2.VideoCapture(rf"{TESTDATA_DIR}/real3.avi")
    detections = []
    frames = []
    while cap.isOpened():
        success, frame = cap.read()
        if not success:
            break
        frames.append(frame)
        detections.append(detection.detect_devices(frame, 220))
    timestamps = file.read_obj(rf"{TESTDATA_DIR}/real3-timestamps.txt")
    sequences = file.read_obj(rf"{TESTDATA_DIR}/real3-sequences.txt")
    tracks = tracking.perform_tracking(detections, timestamps)
    debug.debug_tracking(frames, timestamps, tracks, 1)
    _print_results(sequences, tracks)


def test__process_sequencing__generated_video_100():
    tracking.DEBUG = False
    timestamps = file.read_obj(r"../../tmp-timestamps.txt")
    # noinspection PyArgumentList
    cap = cv2.VideoCapture(r"../testdata/video/test_detection_generated_10x10.avi")
    device_service = DeviceService(None, None)
    devices = device_service._track_devices(cap)
    assert len(devices) == 100


def test__process_sequencing__generated_video_400():
    tracking.DEBUG = False
    # noinspection PyArgumentList
    cap = cv2.VideoCapture(r"../testdata/video/test_detection_generated_20x20.avi")
    device_service = DeviceService(None, None)
    devices = device_service._track_devices(cap)
    assert len(devices) == 400


def test__process_sequencing__live():
    tracking.DEBUG = True
    cap = cv2.VideoCapture(0)
    device_service = DeviceService(None, None)
    devices = device_service._track_devices(cap)


def test__generated():
    filename = rf'{TESTDATA_DIR}/tracking/generated'
    timestamps, positions, sequences, ground_truth = test_generator.generate_video((35, 35), filename, movement=3, max_start_delay=20, device_color_variation=60, fps=10)
    cap = cv2.VideoCapture(f"{filename}.mp4")
    frames = []
    positions = []
    while cap.isOpened():
        success, frame = cap.read()
        if not success:
            break
        frames.append(frame)
        positions.append(detection.detect_devices(frame, 200))
    # timestamps = file.read_obj(rf"{TESTDATA_DIR}/tracking/generated-timestamps.txt")
    # sequences = file.read_obj(rf"{TESTDATA_DIR}/tracking/generated-sequences.txt")
    # positions = file.read_obj(rf"{TESTDATA_DIR}/tracking/generated-positions.txt")
    detected_tracks = tracking.perform_tracking(positions, timestamps)

    frames = []
    hypothesis = {'frames': frames, 'class': 'video'}
    for i, t in enumerate(timestamps):
        hypotheses = []
        frame = {'timestamp': t, 'num': i, 'class': 'frame', 'hypotheses': hypotheses}
        frames.append(frame)
        for idx, track in enumerate(detected_tracks):
            h = {'height': 1, 'width': 1}
            for state in track.history:
                if state.timestamp == t and state.visible:
                    h['x'] = state.position.x
                    h['y'] = state.position.y
                    h['id'] = str(idx)
                    hypotheses.append(h)
                    break
    file.write_obj(f"{filename}-hypothesis.txt", hypothesis)
    evaluator = MOTEvaluation(ground_truth, hypothesis, 0.01)
    evaluator.evaluate()
    evaluator.printResults()
    evaluator.printTrackStatistics()
    # evaluator.getMOTA()
    # evaluator.getMOTP()
    print(evaluator.getRelativeStatistics())
    print(evaluator.getAbsoluteStatistics())


def _print_results(sequences, detected_tracks, actual_tracks=None):
    sequence_length = len(sequences[0])
    missed_devices = []
    mapped_devices = 0
    total_devices = len(sequences)
    mapped_tracks = {}
    incorrect_tracks = []
    if actual_tracks:
        for track in actual_tracks.values():
            mapped_tracks[track[0]] = track

    actual_to_detected = [[] for _ in actual_tracks]

    def get_track(pos, frame_index):
        for i, t in enumerate(actual_tracks):
            if t[frame_index] == pos:
                return i
        return None

    id_switches = 0
    for track in detected_tracks:
        if actual_tracks:
            get(track)

            actual_track = mapped_tracks.get(track.history[0].position, None)
            if actual_track:
                for idx, state in enumerate(track.history):
                    if state.position != actual_track[idx]:
                        incorrect_tracks.append(track)
                        break
        found = False
        for sequence in sequences:
            if track.get_sequence(sequence_length) == sequence:
                found = True
                sequences.remove(sequence)
                break
        if found:
            mapped_devices += 1
        else:
            missed_devices.append(track)
    # for track in missed_devices:
    #     print()
    #     print(f"{track.get_sequence(sequence_length)}, {len(track.get_sequence())}")
    #     print(f"states: {len(track.history)} {track.history}")
    # for sequence in sequences:
    #     print(f"{sequence}, {len(sequence)}")
    print('\nResults')
    print(f"Discovered tracks: {len(detected_tracks)}")
    if actual_tracks:
        print(f"Incorrect tracks: {len(incorrect_tracks)}")
    print(f"False positives: {len(missed_devices)}")
    print(f"False negatives:  {len(sequences)}")
    print(f"Mapped: {mapped_devices} of {total_devices} ({mapped_devices/total_devices*100}%)")
